"""
Contains code to handle detection as well as bigfish wrappers related to spot detection.
"""
from ..hints import pipeline_parameters

from ._preprocess import ParameterInputError
from ._preprocess import check_integrity, convert_parameters_types

from ..gui.napari_visualiser import correct_spots, threshold_selection
from ..gui import add_default_loading
from ..gui import detection_parameters_promt

from ..interface import get_voxel_size
from ..utils import compute_anisotropy_coef
from ._signaltonoise import compute_snr_spots

from magicgui import magicgui

from types import GeneratorType
from napari.types import LayerDataTuple

import numpy as np
from numpy import NaN
import pandas as pd
import FreeSimpleGUI as sg
import bigfish.detection as detection
import bigfish.stack as stack
import bigfish.multistack as multistack
import bigfish.classification as classification
from bigfish.detection.spot_detection import get_object_radius_pixel
from skimage.measure import regionprops
from scipy.ndimage import binary_dilation



def compute_auto_threshold(images, voxel_size=None, spot_radius=None, log_kernel_size=None, minimum_distance=None, im_number= 15, crop_zstack= None) :
    """
    Compute bigfish auto threshold efficiently for list of images. In case on large set of images user can set im_number to only consider a random subset of image for threshold computation.
    """
    # check parameters
    stack.check_parameter(images = (list, np.ndarray, GeneratorType,), voxel_size=(int, float, tuple, list, type(None)),spot_radius=(int, float, tuple, list, type(None)),log_kernel_size=(int, float, tuple, list, type(None)),minimum_distance=(int, float, tuple, list, type(None)), im_number = int, crop_zstack= (type(None), tuple))

    # if one image is provided we enlist it
    if not isinstance(images, list):
        if isinstance(images, np.ndarray) : 
            stack.check_array(images,ndim=[2, 3],dtype=[np.uint8, np.uint16, np.float32, np.float64])
            ndim = images.ndim
            images = [images]
        else : 
            images = [image for image in images]
            for image in images : stack.check_array(image,ndim=[2, 3],dtype=[np.uint8, np.uint16, np.float32, np.float64])
            ndim = images[0].ndim

    else:
        ndim = None
        for i, image in enumerate(images):
            stack.check_array(image,ndim=[2, 3],dtype=[np.uint8, np.uint16, np.float32, np.float64])
            if i == 0:
                ndim = image.ndim
            else:
                if ndim != image.ndim:
                    raise ValueError("Provided images should have the same "
                                     "number of dimensions.")
    if len(images) > im_number : #if true we select a random sample of images
        idx = np.arange(len(images),dtype= int)
        np.random.shuffle(idx)
        images = [images[i] for i in idx]
        
    #Building a giant 3D array containing all information for threshold selection -> cheating detection.automated_threshold_setting that doesn't take lists and doesn't use spatial information.
    if type(crop_zstack) == type(None) :
        crop_zstack = (0, len(images[0]))
    
    log_kernel_size, minimum_distance = _compute_threshold_parameters(ndim, voxel_size, spot_radius, minimum_distance, log_kernel_size)
    images_filtered = np.concatenate(
        [stack.log_filter(image[crop_zstack[0]: crop_zstack[1]], sigma= log_kernel_size) for image in images],
         axis= ndim -1)
    max_masks = np.concatenate(
        [detection.local_maximum_detection(image[crop_zstack[0]: crop_zstack[1]], min_distance= minimum_distance) for image in images],
         axis= ndim -1)
    threshold = detection.automated_threshold_setting(images_filtered, max_masks)

    return threshold

def _compute_threshold_parameters(ndim, voxel_size, spot_radius, minimum_distance, log_kernel_size) :

    # check consistency between parameters - detection with voxel size and
    # spot radius
    if (voxel_size is not None and spot_radius is not None
            and log_kernel_size is None and minimum_distance is None):
        if isinstance(voxel_size, (tuple, list)):
            if len(voxel_size) != ndim:
                raise ValueError("'voxel_size' must be a scalar or a sequence "
                                 "with {0} elements.".format(ndim))
        else:
            voxel_size = (voxel_size,) * ndim
        if isinstance(spot_radius, (tuple, list)):
            if len(spot_radius) != ndim:
                raise ValueError("'spot_radius' must be a scalar or a "
                                 "sequence with {0} elements.".format(ndim))
        else:
            spot_radius = (spot_radius,) * ndim
        log_kernel_size = get_object_radius_pixel(
            voxel_size_nm=voxel_size,
            object_radius_nm=spot_radius,
            ndim=ndim)
        minimum_distance = get_object_radius_pixel(
            voxel_size_nm=voxel_size,
            object_radius_nm=spot_radius,
            ndim=ndim)

    # check consistency between parameters - detection with kernel size and
    # minimal distance
    elif (voxel_size is None and spot_radius is None
          and log_kernel_size is not None and minimum_distance is not None):
        if isinstance(log_kernel_size, (tuple, list)):
            if len(log_kernel_size) != ndim:
                raise ValueError("'log_kernel_size' must be a scalar or a "
                                 "sequence with {0} elements.".format(ndim))
        else:
            log_kernel_size = (log_kernel_size,) * ndim
        if isinstance(minimum_distance, (tuple, list)):
            if len(minimum_distance) != ndim:
                raise ValueError("'minimum_distance' must be a scalar or a "
                                 "sequence with {0} elements.".format(ndim))
        else:
            minimum_distance = (minimum_distance,) * ndim

    # check consistency between parameters - detection in priority with kernel
    # size and minimal distance
    elif (voxel_size is not None and spot_radius is not None
          and log_kernel_size is not None and minimum_distance is not None):
        if isinstance(log_kernel_size, (tuple, list)):
            if len(log_kernel_size) != ndim:
                raise ValueError("'log_kernel_size' must be a scalar or a "
                                 "sequence with {0} elements.".format(ndim))
        else:
            log_kernel_size = (log_kernel_size,) * ndim
        if isinstance(minimum_distance, (tuple, list)):
            if len(minimum_distance) != ndim:
                raise ValueError("'minimum_distance' must be a scalar or a "
                                 "sequence with {0} elements.".format(ndim))
        else:
            minimum_distance = (minimum_distance,) * ndim

    # missing parameters
    else:
        raise ValueError("One of the two pairs of parameters ('voxel_size', "
                         "'spot_radius') or ('log_kernel_size', "
                         "'minimum_distance') should be provided.")
    
    return log_kernel_size, minimum_distance

def cluster_detection(spots, voxel_size, radius = 350, nb_min_spots = 4, keys_to_compute = ["clustered_spots", "clusters"]) :
    """
    Performs `bigfish.detection.cluster_detection()` to detect clusters.
    Then offers possibility to get results sorted in pandas dataframe.

    Parameters
    ----------
        spots : np.ndarray
            Coordinates of the detected spots with shape (nb_spots, 3) or (nb_spots, 2).
        voxel_size : int, float, Tuple(int, float) or List(int, float)
            Size of a voxel, in nanometer. One value per spatial dimension (zyx or yx dimensions). If it's a scalar, the same value is applied to every dimensions.
        radius : int
            The maximum distance between two samples for one to be considered as in the neighborhood of the other. Radius expressed in nanometer.
        nb_min_spots : int
            The number of spots in a neighborhood for a point to be considered as a core point (from which a cluster is expanded). This includes the point itself.
        keys_to_compute : list[str], str
            keys from (clustered_spots, clusters, clustered_spots_dataframe, clusters_dataframe)
                --> clustered_spots : np.ndarray
                    Coordinates of the detected spots with shape (nb_spots, 4) or (nb_spots, 3). One coordinate per dimension (zyx or yx coordinates) plus the index of the cluster assigned to the spot. If no cluster was assigned, value is -1.
                --> clusters : np.ndarray
                    Array with shape (nb_clusters, 5) or (nb_clusters, 4). One coordinate per dimension for the clusters centroid (zyx or yx coordinates), the number of spots detected in the clusters and its index.
                --> clustered_spots_dataframe
                --> clusters_dataframe
    
    Returns
    -------
        res : dict
            keys : keys from `keys_to_compute` argument : (clustered_spots, clusters, clustered_spots_dataframe, clusters_dataframe)    
    """

    if isinstance(keys_to_compute, str) : keys_to_compute = [keys_to_compute]
    elif isinstance(keys_to_compute, list) : pass
    else : raise TypeError("Wrong type for keys_to_compute. Should be list[str] or str. It is {0}".format(type(keys_to_compute)))
    if len(spots) == 0 :
        res = {'clustered_spots' : np.empty(shape=(0,len(voxel_size) + 1), dtype=int), 'clusters' : np.empty(shape=(0,len(voxel_size) + 2), dtype=int), 'clustered_spots_dataframe' : pd.DataFrame(columns= ["id", "cluster_id", "z", "y", "x"]), 'clusters_dataframe' : pd.DataFrame(columns= ["id", "z", "y", "x", "spot_number"])}
        return {key : res[key] for key in keys_to_compute}
    else : res = {}
    voxel_size = tuple([int(d) for d in voxel_size])
    clustered_spots, clusters = detection.detect_clusters(spots, voxel_size= voxel_size, radius= radius, nb_min_spots= nb_min_spots)


    if 'clustered_spots' in keys_to_compute :
        res['clustered_spots'] = clustered_spots
        voxel_size
    if 'clusters' in keys_to_compute : 
        res['clusters'] = clusters

    if 'clustered_spots_dataframe' in keys_to_compute :
        res['clustered_spots_dataframe'] = _compute_clustered_spots_dataframe(clustered_spots)
    
    if 'clusters_dataframe' in keys_to_compute :
        res['clusters_dataframe'] = _compute_cluster_dataframe(clusters)

    return res

def initiate_detection(user_parameters : pipeline_parameters, map_, shape) :
    is_3D_stack= user_parameters['is_3D_stack']
    is_multichannel = user_parameters['is_multichannel']
    do_dense_region_deconvolution = user_parameters['do_dense_regions_deconvolution']
    do_clustering = user_parameters['do_cluster_computation']
    detection_parameters = user_parameters.copy()
    
    #Attempt to read voxel size from metadata
    voxel_size = get_voxel_size(user_parameters['image_path'])
    if voxel_size is None or not user_parameters.get('voxel_size') is None:
        pass
    else :
        detection_parameters['voxel_size'] = [int(v) for v in voxel_size]
        detection_parameters['voxel_size_z'] = round(voxel_size[0])
        detection_parameters['voxel_size_y'] = round(voxel_size[1])
        detection_parameters['voxel_size_x'] = round(voxel_size[2])

    #Setting default spot size to 1.5 voxel
    if detection_parameters.get('spot_size') is None :
        detection_parameters['spot_size_z'] = round(detection_parameters['voxel_size_z']*1.5) if not detection_parameters['voxel_size_z'] is None else None
        detection_parameters['spot_size_y'] = round(detection_parameters['voxel_size_y']*1.5) if not detection_parameters['voxel_size_y'] is None else None
        detection_parameters['spot_size_x'] = round(detection_parameters['voxel_size_x']*1.5) if not detection_parameters['voxel_size_x'] is None else None

    while True :
        detection_parameters = detection_parameters_promt(
            is_3D_stack=is_3D_stack,
            is_multichannel=is_multichannel,
            do_dense_region_deconvolution=do_dense_region_deconvolution,
            do_clustering=do_clustering,
            segmentation_done= user_parameters['segmentation_done'],
            default_dict=detection_parameters
            )
        if type(detection_parameters) == type(None) : return None
        try :
            detection_parameters = convert_parameters_types(detection_parameters)
            detection_parameters = check_integrity(
                detection_parameters, 
                do_dense_region_deconvolution,
                do_clustering, 
                is_multichannel, 
                user_parameters['segmentation_done'], 
                map_, 
                shape
                )
        except ParameterInputError as error:
            sg.popup(error)
        else :
            user_parameters.update(detection_parameters)
            break
    
    return user_parameters

@add_default_loading
def _launch_detection(image, image_input_values: dict) :

    """
    Performs spots detection
    """
    
    #Extract parameters
    voxel_size = image_input_values['voxel_size']
    threshold = image_input_values.get('threshold')
    threshold_penalty = image_input_values.setdefault('threshold penalty', 1)
    spot_size = image_input_values.get('spot_size')
    log_kernel_size = image_input_values.get('log_kernel_size')
    minimum_distance = image_input_values.get('minimum_distance')
    threshold_user_selection = image_input_values['show_interactive_threshold_selector']
    
    if type(threshold) == type(None) :     
        threshold = threshold_penalty * compute_auto_threshold(image, voxel_size=voxel_size, spot_radius=spot_size, log_kernel_size=log_kernel_size, minimum_distance=minimum_distance)
    
    filtered_image = _apply_log_filter(
        image=image,
        voxel_size=voxel_size,
        spot_radius=spot_size,
        log_kernel_size = log_kernel_size,
    )

    local_maxima = _local_maxima_mask(
        image_filtered=filtered_image,
        voxel_size=voxel_size,
        spot_radius=spot_size,
        minimum_distance=minimum_distance
    )

    if threshold_user_selection :

        threshold_slider = _create_threshold_slider(
            logfiltered_image=filtered_image,
            local_maxima=local_maxima,
            default=threshold,
            min_value=filtered_image[local_maxima].min(),
            max_value=filtered_image[local_maxima].max(),
            voxel_size=voxel_size
        )

        spots, threshold = threshold_selection(
            image=image,
            filtered_image=filtered_image,
            threshold_slider=threshold_slider,
            voxel_size=voxel_size
        )
    else :
        spots = detection.spots_thresholding(
            image=filtered_image,
            mask_local_max=local_maxima,
            threshold=threshold
        )[0]
        
    return spots, threshold

@add_default_loading
def launch_dense_region_deconvolution(image, spots, image_input_values: dict,) :
    """
    Performs spot decomposition

    Returns
    -------
        spots : np.ndarray
            Array(nb_spot, dim) (dim is either 3 or 2)
        fov_res : dict
            keys : spot_number, spotsSignal_median, spotsSignal_mean, spotsSignal_std, median_pixel, mean_pixel, snr_median, snr_mean, snr_std, cell_medianbackground_mean, cell_medianbackground_std, cell_meanbackground_mean, cell_meanbackground_std, cell_stdbackground_mean, cell_stdbackground_std
    """
    
    ##Initiate lists
    voxel_size = image_input_values['voxel_size']
    spot_size = image_input_values.get('spot_size')
    ##deconvolution parameters
    alpha = image_input_values.get('alpha')
    beta = image_input_values.get('beta')
    gamma = image_input_values.get('gamma')
    deconvolution_kernel = image_input_values.get('deconvolution_kernel')
    dim = image_input_values['dim']
        
    spots, dense_regions, ref_spot = detection.decompose_dense(image=image, spots=spots, voxel_size=voxel_size, spot_radius=spot_size, kernel_size=deconvolution_kernel, alpha=alpha, beta=beta, gamma=gamma)
    del dense_regions, ref_spot

    return spots

@add_default_loading
def launch_post_detection(image, spots, image_input_values: dict,) :
    fov_res = {}
    dim = image_input_values['dim']
    voxel_size = image_input_values['voxel_size']
    spot_size = image_input_values.get('spot_size')

    #features
    fov_res['spot_number'] = len(spots)
    snr_res = compute_snr_spots(image, spots, voxel_size, spot_size)
    if len(spots) == 0 :
        fov_res['spotsSignal_median'], fov_res['spotsSignal_mean'], fov_res['spotsSignal_std'] = np.NaN, np.NaN, np.NaN
    else :
        if dim == 3 :
            Z,Y,X = list(zip(*spots))
            spots_values = image[Z,Y,X]
        else :
            Y,X = list(zip(*spots))
            spots_values = image[Y,X]
        fov_res['spotsSignal_median'], fov_res['spotsSignal_mean'], fov_res['spotsSignal_std'] = np.median(spots_values), np.mean(spots_values), np.std(spots_values)
    
    fov_res['median_pixel'] = np.median(image)
    fov_res['mean_pixel'] = np.mean(image)

    #appending results
    fov_res.update(snr_res)

    return fov_res

def _compute_cell_snr(image: np.ndarray, bbox, spots, voxel_size, spot_size) :
    
    min_y, min_x, max_y, max_x = bbox
    image= image[min_y: max_y, min_x: max_x]

    if len(spots) == 0 :
        res = {
           'snr_mean' : np.NaN,
           'snr_median' : np.NaN,
           'snr_std' : np.NaN,
        }

        return res

    if len(spots[0]) == 3 :
        Z,Y,X = zip(*spots)
        spots = np.array(
            list(zip(Y,X)),
            dtype= int
        )
        voxel_size = voxel_size[1:]
        spot_size = spot_size[1:]

    snr_dict = compute_snr_spots(
            image= image,
            spots= spots,
            spot_radius= spot_size,
            voxel_size=voxel_size
            )
        
    return snr_dict

@add_default_loading
def launch_cell_extraction(
    acquisition_id, 
    spots, 
    clusters, 
    spots_cluster_id,
    image, 
    nucleus_signal, 
    cell_label, 
    nucleus_label, 
    user_parameters : pipeline_parameters
    ) :

    #Extract parameters
    dim = user_parameters['dim']
    do_clustering = user_parameters['do_cluster_computation']
    voxel_size = user_parameters['voxel_size']

    if do_clustering :
        if len(clusters) > 0 :
            free_spots = spots[spots_cluster_id == -1].astype(int)
            clustered_spots = spots[spots_cluster_id != -1].astype(int)

            other_coords = {
                'clusters_coords' : clusters,
                'clustered_spots' : clustered_spots,
                'free_spots' : free_spots,
                }
        else :
            other_coords = None
    else : other_coords = None
    if do_clustering : do_clustering = len(clusters) > 0

    if image.ndim == 3 :
        image = stack.maximum_projection(image)
    if nucleus_signal.ndim == 3 :
        nucleus_signal = stack.maximum_projection(nucleus_signal)
    
    cells_results = multistack.extract_cell(
        cell_label=cell_label,
        ndim=dim,
        nuc_label=nucleus_label,
        rna_coord=spots,
        others_coord=other_coords,
        image=image
    )

    #BigFish features
    features_names = ['acquisition_id', 'cell_id', 'cell_bbox'] + classification.get_features_name(
        names_features_centrosome=False,
        names_features_area=True,
        names_features_dispersion=True,
        names_features_distance=True,
        names_features_foci=do_clustering,
        names_features_intranuclear=True,
        names_features_protrusion=False,
        names_features_topography=True
    )

    #Nucleus features : area is computed in bigfish
    features_names += ['nucleus_mean_signal', 'nucleus_median_signal', 'nucleus_max_signal', 'nucleus_min_signal']
    features_names += ['snr_mean', 'snr_median', 'snr_std']
    features_names += ['cell_center_coord','foci_number','foci_in_nuc_number']
    features_names += ['rna_coords','cluster_coords', 'clustered_spots_coords', 'free_spots_coords']
    features_names += ['clustered_spot_number', 'free_spot_number']

    result_frame = pd.DataFrame()

    for cell in cells_results :

        #Extract cell results
        cell_id = cell['cell_id']
        cell_mask = cell['cell_mask']
        nuc_mask = cell ['nuc_mask']
        cell_bbox = cell['bbox'] # (min_y, min_x, max_y, max_x)
        min_y, min_x, max_y, max_x = cell['bbox'] # (min_y, min_x, max_y, max_x)
        nuc_signal = nucleus_signal[min_y:max_y, min_x:max_x]
        rna_coords = cell['rna_coord']
        foci_coords = cell.get('clusters_coords')
        clustered_spots_coords = cell.get('clustered_spots')
        free_spots_coords = cell.get('free_spots')
        signal = cell['image']

        if do_clustering :
            if len(clusters) > 0 :
                compute_foci = True
            else : 
                compute_foci = False
        else : compute_foci = False

        with np.errstate(divide= 'ignore', invalid= 'ignore') :
            features = classification.compute_features(
                cell_mask=cell_mask,
                nuc_mask=nuc_mask,
                ndim=dim,
                rna_coord= rna_coords,
                foci_coord=foci_coords,
                voxel_size_yx= float(voxel_size[-1]),
                smfish=signal,
                centrosome_coord=None,
                compute_centrosome=False,
                compute_area=True,
                compute_dispersion=True,
                compute_distance=True,
                compute_foci= compute_foci,
                compute_intranuclear=True,
                compute_protrusion=False,
                compute_topography=True
            )

        #center of cell coordinates
        local_cell_center = regionprops(
            label_image=cell_mask.astype(int)
        )[0]['centroid']
        cell_center = (local_cell_center[0] + min_y, local_cell_center[1] + min_x)

        #foci in nucleus
        if type(foci_coords) != type(None) :
            if len(foci_coords) == 0 : 
                foci_number = 0
                foci_in_nuc_number = 0
            else :
                foci_number = len(foci_coords)
                foci_index = list(zip(*foci_coords))
                if len(foci_index) == 5 :
                    foci_index = foci_index[1:3]
                elif len(foci_index) == 4 :
                    foci_index = foci_index[:2]
                else : raise AssertionError("Impossible number of dim for foci : ", len(foci_index))
                foci_in_nuc_number = nuc_mask[tuple(foci_index)].astype(bool).sum()
        else :
            foci_number = np.NaN
            foci_in_nuc_number = np.NaN

        #Signal to noise
        snr_dict = _compute_cell_snr(
            image,
            cell_bbox,
            spots=rna_coords,
            voxel_size=voxel_size,
            spot_size=user_parameters['spot_size']
        )

        snr_mean = snr_dict['snr_mean']
        snr_median = snr_dict['snr_median']
        snr_std = snr_dict['snr_std']

        features = list(features)
        features += [np.mean(nuc_signal), np.median(nuc_signal), np.max(nuc_signal), np.min(nuc_signal)]
        features += [snr_mean, snr_median, snr_std]
        features += [cell_center, foci_number, foci_in_nuc_number]

        features = [acquisition_id, cell_id, cell_bbox] + features
        features += [rna_coords, foci_coords, clustered_spots_coords, free_spots_coords]
        features += [len(clustered_spots_coords) if type(clustered_spots_coords) != type(None) else None]
        features += [len(free_spots_coords) if type(free_spots_coords) != type(None) else None]
        
        result_frame = pd.concat([
            result_frame,
            pd.DataFrame(columns = features_names, data= (features,)),
        ],
        axis= 0
        )
    
    return result_frame

@add_default_loading
def launch_clustering(spots, user_parameters : pipeline_parameters): 

    voxel_size = user_parameters['voxel_size']
    nb_min_spots = user_parameters['min_number_of_spots']
    cluster_size = user_parameters['cluster_size']

    cluster_result_dict = cluster_detection(
        spots=spots,
        voxel_size=voxel_size,
        radius=cluster_size,
        nb_min_spots=nb_min_spots,
        keys_to_compute= ['clusters','clustered_spots']
    )

    clusters = cluster_result_dict['clusters']
    clustered_spots = cluster_result_dict['clustered_spots']

    return clusters, clustered_spots

def launch_detection(
        image,
        other_image,
        user_parameters : pipeline_parameters,
        cell_label= None,
        nucleus_label = None,
        hide_loading=False,
        ) :
    """
    Main call for features computation :
    --> spot dection
    --> dense regions deconv
    --> cluster (opt)
    --> spot correction
    --> general features computations
    --> cell extractions
    --> cell features computations
    
    RETURNS
    -------
        user_parameters : dict
        spots : np.ndarray
        clusters : np.ndarray

    USER_PARAMETERS UPDATE
    ----------------------
        'threshold'
    """
    fov_result = {}
    do_dense_region_deconvolution = user_parameters['do_dense_regions_deconvolution']
    do_clustering = user_parameters['do_cluster_computation']

    spots, threshold  = _launch_detection(image, user_parameters, hide_loading = hide_loading)
        
    if do_dense_region_deconvolution : 
        spots = launch_dense_region_deconvolution(image, spots, user_parameters, hide_loading = hide_loading)
        
    if do_clustering : 
        clusters, clustered_spots = launch_clustering(spots, user_parameters, hide_loading = hide_loading) #012 are coordinates #3 is number of spots per cluster, #4 is cluster index
        spots, spots_cluster_id = clustered_spots[:,:-1], clustered_spots[:,-1]

    else : 
        clusters = None
        spots_cluster_id = None

    user_parameters['threshold'] = threshold

    if user_parameters['show_napari_corrector'] :

        spots, clusters, new_cluster_radius, new_min_spot_number = correct_spots(
            image, 
            spots, 
            user_parameters['voxel_size'],
            clusters=clusters,
            spot_cluster_id = spots_cluster_id,
            cluster_size= user_parameters.get('cluster_size'),
            min_spot_number= user_parameters.setdefault('min_number_of_spots', 0),
            cell_label=cell_label,
            nucleus_label=nucleus_label,
            other_images=other_image
            )
        
        if type(new_cluster_radius) != type(None) :
            user_parameters['cluster_size'] = new_cluster_radius
        if type(new_min_spot_number) != type(None) :
            user_parameters['min_number_of_spots'] = new_min_spot_number
        
        if do_clustering :
            spots, spots_cluster_id = spots[:,:-1], spots[:,-1]
        else :
            spots_cluster_id = None

    post_detection_dict = launch_post_detection(image, spots, user_parameters, hide_loading = hide_loading)
    fov_result.update(post_detection_dict)
    
    return user_parameters, fov_result, spots, clusters, spots_cluster_id
            

def launch_features_computation(
        acquisition_id, 
        image, 
        nucleus_signal, 
        spots, 
        clusters, 
        spots_cluster_id, 
        nucleus_label, 
        cell_label, 
        user_parameters : pipeline_parameters, 
        frame_results
        ) :

    dim = image.ndim
    if user_parameters['do_cluster_computation'] : 
        frame_results['cluster_number'] = len(clusters)
        if dim == 3 :
            frame_results['total_spots_in_clusters'] = clusters.sum(axis=0)[3] if len(clusters) >0 else  0
        else :
            frame_results['total_spots_in_clusters'] = clusters.sum(axis=0)[2] if len(clusters) >0 else  0
    
    if type(cell_label) != type(None) and type(nucleus_label) != type(None): 
        
        try :
            cell_result_dframe = launch_cell_extraction(
                acquisition_id=acquisition_id,
                spots=spots,
                clusters=clusters,
                spots_cluster_id = spots_cluster_id,
                image=image,
                nucleus_signal=nucleus_signal,
                cell_label= cell_label,
                nucleus_label=nucleus_label,
                user_parameters=user_parameters,
            )

        except IndexError as e: #User loaded a segmentation and no cells can be extracted out of it.
            raise Exception("No cell was fit for quantification in segmentation. This can happen if you loaded empty segmentation or there is a missmatch between cytoplasm and nuclei.\nIf you didn't load segmentation please report the issue as this should not happen.")

    else :
        cell_result_dframe = pd.DataFrame()

    frame_results['acquisition_id'] = acquisition_id
    if type(cell_label) != type(None) and type(nucleus_label) != type(None):
        frame_results['cell_number'] = len(cell_result_dframe) 
    else : 
        frame_results['cell_number'] = NaN
    frame_results['spots'] = spots
    frame_results['clusters'] = clusters
    frame_results['spots_cluster_id'] = spots_cluster_id
    frame_results.update(user_parameters)
    frame_results['threshold'] = user_parameters['threshold']

    frame_results = pd.DataFrame(columns= frame_results.keys(), data= (frame_results.values(),))

    #Adding name column
    result_col = list(frame_results.columns)
    cell_result_col = list(cell_result_dframe.columns)
    name = "acquisition_{0}".format(acquisition_id)
    frame_results['name'] = name
    frame_results = frame_results.loc[:,['name'] + result_col]
    if user_parameters['segmentation_done'] : 
        cell_result_dframe['name'] = name
        cell_result_dframe = cell_result_dframe.loc[:,['name'] + cell_result_col]
        if 'nb_rna_in_nuc' in cell_result_dframe.columns and 'nb_rna_out_nuc' in cell_result_dframe.columns :
            cell_result_dframe['total_rna_number'] = cell_result_dframe['nb_rna_in_nuc'] + cell_result_dframe['nb_rna_out_nuc']
        else : # This can happen when segmentation is performed and detects cells but they are on fov edges and thus removed by big-fish.
            print("\033[1;31m All segmented cells where skipped because they are found on fov edges (incomplete cells), if you want to analyse this image check segmentation.\033[00m")
            cell_result_dframe['nb_rna_in_nuc'] = np.NaN
            cell_result_dframe['nb_rna_out_nuc'] = np.NaN
            cell_result_dframe['total_rna_number'] = np.NaN

        
    return frame_results, cell_result_dframe

def _compute_clustered_spots_dataframe(clustered_spots) :
    if len(clustered_spots) == 0 : return pd.DataFrame(columns= ["id", "cluster_id", "z", "y", "x"])
    z, y ,x, cluster_index = list(zip(*clustered_spots))
    ids = np.arange(len(clustered_spots))

    df = pd.DataFrame({
        "id" : ids
        ,"cluster_id" : cluster_index
        ,"z" : z
        ,"y" : y
        ,"x" : x
    })

    null_idx = df[df['cluster_id'] == -1].index
    df.loc[null_idx, 'cluster_id'] = np.NaN

    return df

def _compute_cluster_dataframe(clusters) :
    if len(clusters) == 0 : return pd.DataFrame(columns= ["id", "z", "y", "x", "spot_number"])
    z, y, x, spots_number, cluster_index = list(zip(*clusters))

    df = pd.DataFrame({
        "id" : cluster_index
        ,"z" : z
        ,"y" : y
        ,"x" : x
        ,"spot_number" : spots_number
    })

    return df

def get_nucleus_signal(image, other_images, user_parameters) :
    if user_parameters['is_multichannel'] :
        rna_signal_channel = user_parameters['channel_to_compute']
        nucleus_signal_channel = user_parameters['nucleus channel signal']
        if type(nucleus_signal_channel) == type(None) :
            return np.zeros(shape=image.shape)

        if rna_signal_channel == nucleus_signal_channel :
            nucleus_signal = image
        
        elif nucleus_signal_channel > rna_signal_channel :
            nucleus_signal_channel -=1
            nucleus_signal = other_images[nucleus_signal_channel]
        
        elif nucleus_signal_channel < rna_signal_channel :
            nucleus_signal = other_images[nucleus_signal_channel]

        return nucleus_signal
    else :
        return image

def _create_threshold_slider(
        logfiltered_image : np.ndarray,
        local_maxima : np.ndarray,
        default : int,
        min_value : int,
        max_value : int,
        voxel_size
) :
    
    if isinstance(default, float) : default = round(default)
    min_value = max(min_value,15) #Security to avoid user put too low threshold and crashes Napari if out of memory.

    @magicgui(
        threshold={'widget_type' : 'Slider', 'value' : default, 'min' : min_value, 'max' : max_value, 'tracking' : True,},
        auto_call=False,
        call_button= "Apply"
    )
    def threshold_slider(threshold: int) -> LayerDataTuple:
        spots = detection.spots_thresholding(
            image=logfiltered_image,
            mask_local_max=local_maxima,
            threshold=threshold
        )[0]

        scale = compute_anisotropy_coef(voxel_size)

        layer_args = {
            'size': 5, 
            'scale' : scale, 
            'face_color' : 'transparent', 
            'edge_color' : 'red', 
            'symbol' : 'disc', 
            'opacity' : 0.7, 
            'blending' : 'translucent', 
            'name': 'single spots',
            'features' : {'threshold' : threshold},
            'visible' : True,
            }
        return (spots, layer_args , 'points')
    return threshold_slider

def _apply_log_filter(
        image: np.ndarray,
        voxel_size : tuple,
        spot_radius : tuple,
        log_kernel_size,

) :
    """
    Apply spot detection steps until local maxima step (just before final threshold).
    Return filtered image.
    """
    
    ndim = image.ndim

    if type(log_kernel_size) == type(None) :
        log_kernel_size = get_object_radius_pixel(
                voxel_size_nm=voxel_size,
                object_radius_nm=spot_radius,
                ndim=ndim)
    
    
    image_filtered = stack.log_filter(image, log_kernel_size)
    
    return image_filtered
    
def _local_maxima_mask(
    image_filtered: np.ndarray,
    voxel_size : tuple,
    spot_radius : tuple,
    minimum_distance

    ) : 

    ndim = image_filtered.ndim

    if type(minimum_distance) == type(None) :
        minimum_distance = get_object_radius_pixel(
            voxel_size_nm=voxel_size,
            object_radius_nm=spot_radius,
            ndim=ndim)
    mask_local_max = detection.local_maximum_detection(image_filtered, minimum_distance)
    
    return mask_local_max.astype(bool)

def output_spot_tiffvisual(channel,spots_list, path_output, dot_size = 3, rescale = True):
    
    """
    Outputs a tiff image with one channel being {channel} and the other a mask containing dots where sports are located.
    
    Parameters
    ----------
        channel : np.ndarray
            3D monochannel image
        spots : list[np.ndarray] or np.ndarray
            Spots arrays are ndarray where each element corresponds is a tuple(z,y,x) corresponding to 3D coordinate of a spot
            To plot different spots on different channels a list of spots ndarray can be passed. 
        path_output : str
        dot_size : int
            in pixels
    """
    
    stack.check_parameter(channel = (np.ndarray), spots_list= (list, np.ndarray), path_output = (str), dot_size = (int))
    stack.check_array(channel, ndim= [2,3])
    if isinstance(spots_list, np.ndarray) : spots_list = [spots_list]

    if channel.ndim == 3 : 
        channel = stack.maximum_projection(channel)

    im = np.zeros([1 + len(spots_list)] + list(channel.shape))
    im[0,:,:] = channel

    for level in range(len(spots_list)) :
        if len(spots_list[level]) == 0 : continue
        else :
            spots_mask = np.zeros_like(channel)
            
            #Unpacking spots
            if len(spots_list[level][0]) == 2 :
                Y,X = zip(*spots_list[level])
            elif len(spots_list[level][0]) == 3 :
                Z,Y,X = zip(*spots_list[level])
                del Z
            else :
                Z,Y,X,*_ = zip(*spots_list[level])
                del Z,_
            
            #Reconstructing signal
            spots_mask[Y,X] = 1
            if dot_size > 1 : spots_mask = binary_dilation(spots_mask, iterations= dot_size-1)
            spots_mask = stack.rescale(np.array(spots_mask, dtype = channel.dtype))
            im[level + 1] = spots_mask

    if rescale : channel = stack.rescale(channel, channel_to_stretch= 0)
    stack.save_image(im, path_output, extension= 'tif')


"""
Submodule keeping necessary calls from main pipeline for batch processing.
"""

import os
import pandas as pd
import FreeSimpleGUI as sg

from ..hints import pipeline_parameters

from .input import open_image
from ..interface import write_results
from ..pipeline import reorder_shape, reorder_image_stack, prepare_image_detection
from ..pipeline import cell_segmentation, launch_detection, launch_features_computation
from ..pipeline import launch_spots_extraction
from ..pipeline import get_nucleus_signal
from ..pipeline import _cast_segmentation_parameters, convert_parameters_types
from ..pipeline import plot_segmentation, output_spot_tiffvisual
from ..utils import get_datetime

def window_print(window: sg.Window, *args) :
    print(*args)
    window.refresh()

def batch_pipeline(
        batch_window : sg.Window,
        batch_progress_bar : sg.ProgressBar,
        progress_count : sg.Text,
        parameters : pipeline_parameters,
        filenames_list : list,
        do_segmentation : bool,
        map_ : dict,
        results_df : pd.DataFrame,
        cell_results_df : pd.DataFrame,
        is_3D,
        last_acquisition_id=0,
) :
    
    #Extracting parameters
    input_path = parameters['Batch_folder']
    output_path = parameters['output_folder']
    batch_name = parameters['batch_name']
    time = '_' + get_datetime()

    #Preparing folder
    window_print(batch_window,"Creating folders for output...")
    main_dir = output_path + "/" + batch_name + time + "/"
    os.makedirs(main_dir + "results/", exist_ok=True)
    if parameters['save segmentation'] : os.makedirs(main_dir + "segmentation/", exist_ok=True)
    if parameters['save detection'] : os.makedirs(main_dir + "detection/", exist_ok=True)
    if parameters['extract spots'] : os.makedirs(main_dir + "results/spots_extraction", exist_ok=True)

    #Setting spot detection dimension
    parameters['dim'] = 3 if is_3D else 2

    #Pipeline loop
    window_print(batch_window,"Launching batch analysis...")
    batch_progress_bar.update(max=len(filenames_list))
    filenames_list.sort()
    

    error_log = open(main_dir + "error_log", mode='w')
    error_count = 0

    for acquisition_id, file in enumerate(filenames_list) :
        try :

            #GUI
            window_print(batch_window,"\nNext file : {0}".format(file))
            batch_progress_bar.update(current_count= acquisition_id, max= len(filenames_list))
            progress_count.update(value=str(acquisition_id))
            batch_window = batch_window.refresh()

            #0. Open image
            image = open_image(input_path + '/' + file)
            parameters['image'] = image
            parameters['filename'] = file
            for key_to_clean in [0,2] : 
                if key_to_clean in parameters : del parameters[key_to_clean]

            #1. Re-order shape
            shape = image.shape
            parameters['shape'] = shape
            parameters['reordered_shape'] = reorder_shape(shape, map_=map_)

            #2. Segmentation (opt)
            if do_segmentation :
                window_print(batch_window,"Segmenting cells...")
                im_seg = reorder_image_stack(map_, image)
                parameters = _cast_segmentation_parameters(parameters)
                cytoplasm_label, nucleus_label = cell_segmentation(
                    im_seg,
                    cyto_model_name= parameters['cyto_model_name'],
                    cyto_diameter= parameters['cytoplasm diameter'],
                    nucleus_model_name= parameters['nucleus_model_name'],
                    nucleus_diameter= parameters['nucleus diameter'],
                    channels=[parameters['cytoplasm channel'], parameters['nucleus channel']],
                    do_only_nuc=parameters['Segment only nuclei']
                    )

                parameters['segmentation_done'] = True

                if cytoplasm_label.max() == 0 : #No cell segmented
                    window_print(batch_window,"No cell was segmented, computing next image.")
                    continue
                else : 
                    window_print(batch_window, "{0} cells segmented.".format(cytoplasm_label.max()))

                    if parameters['save segmentation'] :
                        plot_segmentation(
                            cyto_image=im_seg[parameters['cytoplasm channel']],
                            cyto_label= cytoplasm_label,
                            nuc_image= im_seg[parameters['nucleus channel']],
                            nuc_label=nucleus_label,
                            path= main_dir + "segmentation/" + file,
                            do_only_nuc= parameters['Segment only nuclei'],
                        )

            else :
                cytoplasm_label, nucleus_label = None,None
                parameters['segmentation_done'] = False

            #3. Detection, deconvolution, clusterisation
            window_print(batch_window,"Detecting spots...")
            parameters = convert_parameters_types(parameters)
            image, other_image = prepare_image_detection(map_, parameters) 
            nucleus_signal = get_nucleus_signal(image, other_image, parameters)
            try : # Catch error raised if user enter a spot size too small compare to voxel size
                parameters, frame_result, spots, clusters, spot_cluster_id = launch_detection(
                    image,
                    other_image,
                    parameters,
                    cell_label=cytoplasm_label,
                    nucleus_label=nucleus_label,
                    hide_loading=True,
                )

            except ValueError as error :
                if "The array should have an upper bound of 1" in str(error) :
                    window_print(batch_window,"Spot size too small for current voxel size.")
                    continue
                else :
                    raise(error)

            if parameters['save detection'] :
                if parameters['do_cluster_computation'] : 
                    if len(clusters) > 0 :
                        spots_list = [spots, clusters[:,:-2]]
                    else : spots_list = [spots]
                else : spots_list = [spots]
                output_spot_tiffvisual(
                    image,
                    spots_list= spots_list,
                    dot_size=2,
                    path_output= main_dir + "detection/" + file + "_spot_detection.tiff"
                )

            #4. Spots extraction
            window_print(batch_window,"Extracting spots : ")
            if parameters['extract spots'] :

                #Setting parameter for call to lauch spot extraction
                #Only spots have one file per image to avoir memory overload
                parameters['do_spots_excel'] = parameters['xlsx']
                parameters['do_spots_csv'] = parameters['csv']
                parameters['do_spots_feather'] = parameters['feather']
                parameters['spots_filename'] = "spots_extractions_{0}".format(file)
                parameters['spots_extraction_folder'] = main_dir + "results/spots_extraction/"

                launch_spots_extraction(
                        acquisition_id=acquisition_id + last_acquisition_id,
                        user_parameters=parameters,
                        image=image,
                        spots=spots,
                        cluster_id=spot_cluster_id,
                        nucleus_label= nucleus_label,
                        cell_label= cytoplasm_label,
                    )

            #5. Features computation
            window_print(batch_window,"computing features...")
            new_results_df, new_cell_results_df = launch_features_computation(
            acquisition_id=acquisition_id + last_acquisition_id,
            image=image,
            nucleus_signal = nucleus_signal,
            spots=spots,
            clusters=clusters,
            spots_cluster_id=spot_cluster_id,
            nucleus_label = nucleus_label,
            cell_label= cytoplasm_label,
            user_parameters=parameters,
            frame_results=frame_result,
            )

            results_df = pd.concat([
                results_df.reset_index(drop=True), new_results_df.reset_index(drop=True)
            ], axis=0)

            cell_results_df = pd.concat([
                cell_results_df.reset_index(drop=True), new_cell_results_df.reset_index(drop=True)
            ], axis=0)


            #6. Saving results
            window_print(batch_window,"saving image_results...")
            #1 file per batch + 1 file per batch if segmentation
            acquisition_success = write_results(
                results_df, 
                path= main_dir + "results/", 
                filename=batch_name, 
                do_excel= parameters["xlsx"], 
                do_feather= parameters["feather"], 
                do_csv= parameters["csv"],
                overwrite=True,
                )

            if do_segmentation :
                cell_success = write_results(
                    cell_results_df, 
                    path= main_dir + "results/", 
                    filename=batch_name + '_cell_result', 
                    do_excel= parameters["xlsx"], 
                    do_feather= parameters["feather"], 
                    do_csv= parameters["csv"],
                    overwrite=True,
                    )

            window_print(batch_window,"Sucessfully saved.")

        except Exception as error :
            
            error_count +=1
            print("Exception raised for acquisition, writting error in error log.")

            log = [
                f"Error raised during acquisition {acquisition_id}.\n"
            ]
            log.append(str(error) + '\n')

            error_log.writelines(log)

            print("Ignoring current acquisition and proceeding to next one.")
            continue

    
        
    batch_progress_bar.update(current_count= acquisition_id+1, max= len(filenames_list))
    progress_count.update(value=str(acquisition_id+1))
    batch_window = batch_window.refresh()

    if error_count > 0 :
        sg.popup(f"Batch processing finished but {error_count} acquisitions were skipped during quantification.\nFor more informations check error_log in result folder.")

    return results_df, cell_results_df, acquisition_id



import FreeSimpleGUI as sg
import pandas as pd
import os
import numpy as np
from typing import Literal, Union, Any
from .layout import path_layout, parameters_layout, bool_layout, tuple_layout, combo_elmt, add_header, path_layout, radio_layout
from ..interface import open_image, check_format, FormatError


def prompt(layout, add_ok_cancel=True, timeout=None, timeout_key='TIMEOUT_KEY', add_scrollbar=True) :
    """
    Default event : 'Ok', 'Cancel'
    """
    if add_ok_cancel : layout += [[sg.Button('Ok', bind_return_key=True), sg.Button('Cancel')]]

    if add_scrollbar :
        size = (400,500)
        col_elmt = sg.Column(layout, scrollable=True, vertical_scroll_only=True, size=size)
        layout = [[col_elmt]]
    else :
        size = (None,None)
    
    window = sg.Window('small fish', layout=layout, margins=(10,10), size=size, resizable=True, location=None)
    event, values = window.read(timeout=timeout, timeout_key=timeout_key)
    if event == None : 
        window.close()
        quit()

    elif event == 'Cancel' :
        window.close()
        return event,{}
    else : 
        window.close()
        return event, values



def input_image_prompt(
        is_3D_stack_preset=False,
        multichannel_preset = False,
        do_dense_regions_deconvolution_preset= False,
        do_clustering_preset = False,
        do_Napari_correction= False,
) :
    """
        Keys :
        - 'image_path'
        - 'is_3D_stack'
        - 'time stack'
        - 'is_multichannel'
        - 'do_dense_regions_deconvolution'
        - 'Segmentation'
        - 'show_napari_corrector'

    Returns Values

    """
    layout_image_path = path_layout(['image_path'], header= "Image")
    layout_image_path += bool_layout(['3D stack', 'Multichannel stack'],keys= ['is_3D_stack', 'is_multichannel'], preset= [is_3D_stack_preset, multichannel_preset])
    
    if type(do_dense_regions_deconvolution_preset) != type(None) and type(do_clustering_preset) != type(None) and type(do_Napari_correction) != type(None): 
        layout_image_path += bool_layout(['Dense regions deconvolution', 'Compute clusters', 'Open results in Napari'], keys = ['do_dense_regions_deconvolution', 'do_cluster_computation', 'show_napari_corrector'], preset= [do_dense_regions_deconvolution_preset, do_clustering_preset, do_Napari_correction], header= "Pipeline settings")
    
    event, values = prompt(layout_image_path, add_scrollbar=False)

    if event == 'Cancel' :
        return None

    im_path = values['image_path']
    is_3D_stack = values['is_3D_stack']
    is_multichannel = values['is_multichannel']
    
    try :
        image = open_image(im_path)
        check_format(image, is_3D_stack, is_multichannel)
        values.update({'image' : image})
    except FormatError as error:
        sg.popup("Inconsistency between image format and options selected.\n Image shape : {0}".format(image.shape))
    except OSError as error :
        sg.popup('Image format not supported.')
    except ValueError as error :
        sg.popup('Image format not supported.')


    return values

def output_image_prompt(filename) :
    while True :
        relaunch = False
        layout = path_layout(['folder'], look_for_dir= True, header= "Output parameters :")
        layout += parameters_layout(["filename"], default_values= [filename + "_quantification"], size=25)
        layout += bool_layout(['csv','Excel', 'Feather'])
        layout.append([sg.Button('Cancel')])

        event,values= prompt(layout)

        values['filename'] = values['filename'].replace(".xlsx","")
        values['filename'] = values['filename'].replace(".feather","")
        excel_filename = values['filename'] + ".xlsx"
        feather_filename = values['filename'] + ".feather"

        if not values['Excel'] and not values['Feather'] and not values['csv'] :
            sg.popup("Please check at least one box : Excel/Feather/csv")
            relaunch = True
        elif not os.path.isdir(values['folder']) :
            sg.popup("Incorrect folder")
            relaunch = True
        elif os.path.isfile(values['folder'] + excel_filename) and values['Excel']:
            if ask_replace_file(excel_filename) :
                pass
            else :
                relaunch = True
        elif os.path.isfile(values['folder'] + feather_filename) and values['Feather']:
            if ask_replace_file(feather_filename) :
                pass
            else :
                relaunch = True

        if not relaunch : break

    if event == ('Cancel') : return None

    else : return values

def detection_parameters_promt(
        is_3D_stack, 
        is_multichannel, 
        do_dense_region_deconvolution, 
        do_clustering, 
        segmentation_done, 
        default_dict: dict
        ):
    """

    Returns Values
        
    """
    if is_3D_stack : dim = 3
    else : dim = 2

    #Detection
    detection_parameters = ['threshold', 'threshold penalty']
    default_detection = [default_dict.setdefault('threshold',''), default_dict.setdefault('threshold penalty', '1')]
    opt= [True, True]
    if is_multichannel : 
        detection_parameters += ['channel_to_compute']
        opt += [False]
        default_detection += [default_dict.setdefault('channel_to_compute', '')]
    layout = [[sg.Text("Green parameters", text_color= 'green'), sg.Text(" are optional parameters.")]]
    layout += parameters_layout(detection_parameters, header= 'Detection', opt=opt, default_values=default_detection)
    
    if dim == 2 : tuple_shape = ('y','x')
    else : tuple_shape = ('z','y','x')
    opt = {'voxel_size' : False, 'spot_size' : False, 'log_kernel_size' : True, 'minimum_distance' : True}
    unit = {'voxel_size' : 'nm', 'minimum_distance' : 'nm', 'spot_size' : 'radius(nm)', 'log_kernel_size' : 'px'}

    layout += tuple_layout(opt=opt, unit=unit, default_dict=default_dict, voxel_size= tuple_shape, spot_size= tuple_shape, log_kernel_size= tuple_shape, minimum_distance= tuple_shape)

    #Deconvolution
    if do_dense_region_deconvolution :
        default_dense_regions_deconvolution = [default_dict.setdefault('alpha',0.5), default_dict.setdefault('beta',1)]
        layout += parameters_layout(['alpha', 'beta',], default_values= default_dense_regions_deconvolution, header= 'do_dense_regions_deconvolution')
        layout += parameters_layout(['gamma'], unit= 'px', default_values= [default_dict.setdefault('gamma',5)])
        layout += tuple_layout(opt= {"deconvolution_kernel" : True}, unit= {"deconvolution_kernel" : 'px'}, default_dict=default_dict, deconvolution_kernel = tuple_shape)
    
    #Clustering
    if do_clustering :
        layout += parameters_layout(['cluster_size'], unit="radius(nm)", default_values=[default_dict.setdefault('cluster_size',400)])
        layout += parameters_layout(['min_number_of_spots'], default_values=[default_dict.setdefault('min_number_of_spots', 5)])

    if is_multichannel and segmentation_done :
        default_segmentation = [default_dict.setdefault('nucleus channel signal', default_dict.setdefault('nucleus channel',0))]
        layout += parameters_layout(['nucleus channel signal'], default_values=default_segmentation) + [[sg.Text(" channel from which signal will be measured for nucleus features.")]]

    layout += bool_layout(['Interactive threshold selector'], keys=['show_interactive_threshold_selector'], preset=[False])
    layout += path_layout(
        keys=['spots_extraction_folder'],
        look_for_dir=True,
        header= "Individual spot extraction",
        preset= default_dict.setdefault('spots_extraction_folder', '')
    )
    default_filename = default_dict.setdefault("filename","") + "_spot_extraction"
    layout += parameters_layout(
        parameters=['spots_filename'],
        default_values=[default_filename],
        size= 13
    )
    layout += bool_layout(
        ['.csv','.excel','.feather'],
        keys= ['do_spots_csv', 'do_spots_excel', 'do_spots_feather'],
        preset= [default_dict.setdefault('do_spots_csv',False), default_dict.setdefault('do_spots_excel',False),default_dict.setdefault('do_spots_feather',False)]
    )

    event, values = prompt(layout)
    if event == 'Cancel' : return None
    if is_3D_stack : values['dim'] = 3
    else : values['dim'] = 2
    return values

def ask_replace_file(filename:str) :
    layout = [
        [sg.Text("{0} already exists, replace ?")],
        [sg.Button('Yes'), sg.Button('No')]
    ]

    event, values = prompt(layout, add_ok_cancel= False)

    return event == 'Yes'

def ask_cancel_segmentation() :
    layout = [
        [sg.Text("Cancel segmentation ?")],
        [sg.Button('Yes'), sg.Button('No')]
    ]

    event, values = prompt(layout, add_ok_cancel= False)

    return event == 'Yes'

def ask_quit_small_fish() :
    layout = [
        [sg.Text("Quit small fish ?")],
        [sg.Button('Yes'), sg.Button('No')]
    ]

    event, values = prompt(layout, add_ok_cancel= False)

    return event == 'Yes'

def _error_popup(error:Exception) :
    sg.popup('Error : ' + str(error))
    raise error

def _warning_popup(warning:str) :
    sg.popup('Warning : ' + warning)

def _sumup_df(results: pd.DataFrame) :

    COLUMNS = ['acquisition_id','name','threshold', 'spot_number', 'cell_number', 'filename', 'channel_to_compute']

    if len(results) > 0 :
        if 'channel_to_compute' not in results : results['channel_to_compute'] = np.NaN
        res = results.loc[:,COLUMNS]
    else :
        res = pd.DataFrame(columns= COLUMNS)

    return res

def hub_prompt(fov_results : pd.DataFrame, do_segmentation=False) -> 'Union[Literal["Add detection", "Compute colocalisation", "Batch detection", "Rename acquisition", "Save results", "Delete acquisitions", "Reset segmentation", "Reset results", "Segment cells"], dict[Literal["result_table", ""]]]':

    sumup_df = _sumup_df(fov_results)
    
    if do_segmentation :
        segmentation_object = sg.Text('Segmentation in memory', font='8', text_color= 'green')
    else :
        segmentation_object = sg.Text('No segmentation in memory', font='8', text_color= 'red')

    layout = [
        [sg.Text('RESULTS', font= 'bold 13')],
        [sg.Table(values= list(sumup_df.values), headings= list(sumup_df.columns), row_height=20, num_rows= 5, vertical_scroll_only=False, key= "result_table"), segmentation_object],
        [sg.Button('Segment cells'), sg.Button('Add detection'), sg.Button('Compute colocalisation'), sg.Button('Batch detection')],
        [sg.Button('Save results', button_color= 'green'), sg.Button('Save segmentation', button_color= 'green'), sg.Button('Load segmentation', button_color= 'green')],
        [sg.Button('Rename acquisition', button_color= 'gray'), sg.Button('Delete acquisitions',button_color= 'gray'), sg.Button('Reset segmentation',button_color= 'gray'), sg.Button('Reset all',button_color= 'gray'), sg.Button('Open wiki',button_color= 'yellow', key='wiki')],
    ]

    window = sg.Window('small fish', layout= layout, margins= (10,10), location=None)

    while True : 
        event, values = window.read()
        if event == None : quit()
        elif event == 'Help' : pass
        else : 
            window.close()
            return event, values

def coloc_prompt() :
    layout = parameters_layout(['colocalisation distance'], unit= 'nm', header= 'Colocalisation', default_values= 0)
    event, values = prompt(layout)

    if event == 'Ok' :
        return values['colocalisation distance']
    else : return False

def rename_prompt() :
    layout = parameters_layout(['name'], header= "Rename acquisitions", size=12)
    event, values = prompt(layout)
    if event == 'Ok' :
        return values['name']
    else : return False

def ask_detection_confirmation(used_threshold) :
    layout = [
        [sg.Text("Proceed with current detection ?", font= 'bold 10')],
        [sg.Text("Threshold : {0}".format(used_threshold))],
        [sg.Button("Ok"), sg.Button("Restart detection")]
    ]

    event, value = prompt(layout, add_ok_cancel=False, add_scrollbar=False)

    if event == 'Restart detection' :
        return False
    else :
        return True
    
def ask_cancel_detection() :
    layout =[
        [sg.Text("Cancel new detection and return to main window ?", font= 'bold 10')],
        [sg.Button("Yes"), sg.Button("No")]
    ]

    event, value = prompt(layout, add_ok_cancel=False, add_scrollbar=False)

    if event == 'No' :
        return False
    else :
        return True

def ask_confirmation(question_displayed : str) :
    layout =[
        [sg.Text(question_displayed, font= 'bold 10')],
        [sg.Button("Yes"), sg.Button("No")]
    ]

    event, value = prompt(layout, add_ok_cancel=False, add_scrollbar=False)

    if event == 'No' :
        return False
    else :
        return True
    
def prompt_save_segmentation() -> 'dict[Literal["folder","filename","ext"]]':
    while True :
        relaunch = False
        layout = path_layout(['folder'], look_for_dir= True, header= "Output parameters :")
        layout += parameters_layout(["filename"], default_values= ["small_fish_segmentation"], size=25)
        layout += radio_layout(['npy','npz_uncompressed', 'npz_compressed'], key= 'ext')

        event,values= prompt(layout)
        if event == ('Cancel') : 
            return None

        values['filename'] = values['filename'].replace(".npy","")
        values['filename'] = values['filename'].replace(".npz","")
        filename = values['filename']

        if not os.path.isdir(values['folder']) :
            sg.popup("Incorrect folder")
            relaunch = True
        elif os.path.isfile(values['folder'] + filename):
            if ask_replace_file(filename) :
                pass
            else :
                relaunch = True

        if not relaunch : break

    return values

def prompt_load_segmentation() -> 'dict[Literal["nucleus","cytoplasm"]]':
    while True :
        relaunch = False
        layout = path_layout(['nucleus'], look_for_dir= False, header= "Load segmentation :")
        layout += path_layout(['cytoplasm'], look_for_dir= False)

        event,values= prompt(layout)
        if event == ('Cancel') : 
            return None
        
        if not os.path.isfile(values['nucleus']) :
            sg.popup("Incorrect nucleus file selected.")
            relaunch = True

        if not os.path.isfile(values['cytoplasm']) and values['cytoplasm'] != "" :
            sg.popup("Incorrect cytoplasm file selected.")
            relaunch = True
                

        if not relaunch : break


    return values

def prompt_restore_main_menu() -> bool :
    """
    Warn user that software will try to go back to main menu while saving parameters, and propose to save results and quit if stuck.

    Returns True if user want to save and quit else False, to raise error close window.
    """


    layout = [
        [sg.Text("An error was caught while proceeding.\nSoftware can try to save parameters and return to main menu or save results and quit.")],
        [sg.Button("Return to main menu", key='menu'), sg.Button("Save and quit", key='save')]
    ]

    window = sg.Window('small fish', layout=layout, margins=(10,10), auto_size_text=True, resizable=True)
    event, values = window.read(close=True)

    if event is None :
        return None
    elif event == "save" :
        return True
    elif event == "menu" :
        return False
    else :
        raise AssertionError("Unforseen answer")



"""
This script is called when software starts; it is the main loop.
"""
import traceback, os
from small_fish_gui import __version__

import pandas as pd
import FreeSimpleGUI as sg

from .actions import add_detection 
from . actions import save_results 
from . actions import compute_colocalisation 
from . actions import delete_acquisitions, rename_acquisitions 
from . actions import save_segmentation, load_segmentation, segment_cells
from .actions import open_wiki

from ._preprocess import clean_unused_parameters_cache
from ..batch import batch_promp
from ..gui import hub_prompt, prompt_restore_main_menu
from ..hints import pipeline_parameters

#'Global' parameters
user_parameters = pipeline_parameters({'segmentation_done' : False}) #TypedDict
acquisition_id = -1
result_df = pd.DataFrame(columns=['acquisition_id'])
cell_result_df = pd.DataFrame(columns=['acquisition_id'])
global_coloc_df = pd.DataFrame()
cell_coloc_df = dict()
cytoplasm_label = None
nucleus_label = None

while True : #Break this loop to close small_fish
    try :
        result_df = result_df.reset_index(drop=True)
        event, values = hub_prompt(result_df, user_parameters['segmentation_done'])

        if event == 'Add detection' :
            user_parameters = clean_unused_parameters_cache(user_parameters)

            new_result_df, new_cell_result_df, acquisition_id, user_parameters =  add_detection(
                user_parameters=user_parameters,
                acquisition_id=acquisition_id,
                cytoplasm_label = cytoplasm_label,
                nucleus_label = nucleus_label,
                )
            result_df = pd.concat([result_df, new_result_df], axis=0).reset_index(drop=True)
            cell_result_df = pd.concat([cell_result_df, new_cell_result_df], axis=0).reset_index(drop=True)

        elif event == 'Segment cells' :
            nucleus_label, cytoplasm_label, user_parameters = segment_cells(
                user_parameters=user_parameters,
                nucleus_label=nucleus_label,
                cytoplasm_label = cytoplasm_label,
            )

        elif event == 'Save results' :
            save_results(
                result_df=result_df,
                cell_result_df=cell_result_df,
                global_coloc_df=global_coloc_df,
                cell_coloc_df = cell_coloc_df,
            )
        
        elif event == 'Save segmentation' :
            save_segmentation(
                nucleus_label=nucleus_label,
                cytoplasm_label=cytoplasm_label,
            )

        elif event == 'Load segmentation' :
            nucleus_label, cytoplasm_label, user_parameters['segmentation_done'] = load_segmentation(nucleus_label, cytoplasm_label, user_parameters['segmentation_done'])
        
        elif event == 'Compute colocalisation' :
            result_tables = values.setdefault('result_table', []) #Contains the lines selected by the user on the sum-up array.

            global_coloc_df, cell_coloc_df = compute_colocalisation(
                result_tables,
                result_dataframe=result_df,
                cell_result_dataframe=cell_result_df,
                global_coloc_df=global_coloc_df,
                cell_coloc_df=cell_coloc_df,
            )

        elif event == "Reset all" :
            result_df = pd.DataFrame(columns=['acquisition_id'])
            cell_result_df = pd.DataFrame(columns=['acquisition_id'])
            global_coloc_df = pd.DataFrame()
            cell_coloc_df = dict()
            acquisition_id = -1
            user_parameters['segmentation_done'] = False
            cytoplasm_label = None
            nucleus_label = None

        elif event == "Reset segmentation" :
            user_parameters['segmentation_done'] = False
            cytoplasm_label = None
            nucleus_label = None

        elif event == "Delete acquisitions" :
            selected_acquisitions = values.setdefault('result_table', []) #Contains the lines selected by the user on the sum-up array.
            result_df, cell_result_df, global_coloc_df, cell_coloc_df = delete_acquisitions(selected_acquisitions, result_df, cell_result_df, global_coloc_df, cell_coloc_df)

        elif event == "Batch detection" :
            result_df, cell_result_df, acquisition_id, user_parameters, user_parameters['segmentation_done'], cytoplasm_label,nucleus_label = batch_promp(
                result_df,
                cell_result_df,
                acquisition_id=acquisition_id,
                preset=user_parameters,
            )
        
        elif event == "Rename acquisition" :
            selected_acquisitions = values.setdefault('result_table', []) #Contains the lines selected by the user on the sum-up array.
            result_df, cell_result_df, global_coloc_df, cell_coloc_df = rename_acquisitions(selected_acquisitions, result_df, cell_result_df, global_coloc_df, cell_coloc_df)

        elif event == "wiki" :
            open_wiki()

        else :
            break
        

    except Exception as error :
        sg.popup(str(error))

        with open("error_log.txt",'a') as error_log :
            error_log.writelines([
                f"version {__version__}",
                f"error : {error}",
                f"traceback :\n{traceback.format_exc()}",
            ])

        print(f"error_log saved at {os.getcwd()}/error_log.txt. Please consider reporting this by opening an issue on github.")

        save_quit = prompt_restore_main_menu()

        if save_quit is None :
            raise error

        elif save_quit :
            save_results(
                result_df=result_df,
                cell_result_df=cell_result_df,
                global_coloc_df=global_coloc_df,
                cell_coloc_df = cell_coloc_df,
            )
            quit()
        else :
            continue

quit()


"""
Contains Napari wrappers to visualise and correct spots/clusters.
"""

import numpy as np
import napari

from magicgui import widgets

from bigfish.stack import check_parameter
from ._napari_widgets import CellLabelEraser, SegmentationReseter, ChangesPropagater, FreeLabelPicker
from ._napari_widgets import ClusterIDSetter, ClusterMerger, ClusterUpdater, ClusterCreator
from ._napari_widgets import initialize_all_cluster_wizards
from ..utils import compute_anisotropy_coef

def correct_spots(
        image, 
        spots, 
        voxel_size= (1,1,1), 
        clusters= None, 
        spot_cluster_id= None, 
        cluster_size=None, 
        min_spot_number=0, 
        cell_label= None, 
        nucleus_label= None, 
        other_images =[]
        ):
    """
    Open Napari viewer for user to visualize and corrects spots, clusters.

    Returns
    -------
        new_spots,new_clusters
    """
    check_parameter(image= np.ndarray, voxel_size= (tuple,list))
    dim = len(voxel_size)

    if dim == 3 and type(cell_label) != type(None) :
        cell_label = np.repeat(
            cell_label[np.newaxis],
            repeats= len(image),
            axis=0
        )
    if dim == 3 and type(nucleus_label) != type(None) :
        nucleus_label = np.repeat(
            nucleus_label[np.newaxis],
            repeats= len(image),
            axis=0
        )

    scale = compute_anisotropy_coef(voxel_size)
    Viewer = napari.Viewer(ndisplay=2, title= 'Spot correction', axis_labels=['z','y','x'], show= False)
    Viewer.add_image(image, scale=scale, name= "rna signal", blending= 'additive', colormap='red', contrast_limits=[image.min(), image.max()])
    other_colors = ['green', 'blue', 'gray', 'cyan', 'bop orange', 'bop purple'] * ((len(other_images)-1 // 7) + 1)
    for im, color in zip(other_images, other_colors) :  
        Viewer.add_image(im, scale=scale, blending='additive', visible=False, colormap=color, contrast_limits=[im.min(), im.max()])

    single_layer = Viewer.add_points(  # single molecule spots; this layer can be update by user.
        spots, 
        size = 5, 
        scale=scale, 
        face_color= 'transparent',
        border_color ='red', 
        opacity= 1, 
        symbol= 'disc', 
        name= 'single spots',
        features={
            "cluster_id" : spot_cluster_id if not spot_cluster_id is None else [],
            "end" : [True] * len(spots)
            }
        )
    
    if type(clusters) != type(None) :
        if len(clusters) > 0 :
            clusters_coordinates = clusters[:, :dim]
        else :
            clusters_coordinates = np.empty(shape=(0,dim), dtype=int) 
        cluster_layer = Viewer.add_points( # cluster; this layer can be update by user.
            clusters_coordinates, 
            size = 10, 
            scale=scale, 
            face_color= 'blue', 
            opacity= 0.7, 
            symbol= 'diamond', 
            name= 'foci', 
            features= {
                "spot_number" : clusters[:,dim], 
                "cluster_id" : clusters[:,dim+1],
                "end" : [True] * len(clusters_coordinates)
                }, 
            feature_defaults= {"spot_number" : 0, "cluster_id" : -2, "end" : True} # napari features default will not work with np.NaN passing -2 instead.
        )

    if type(cell_label) != type(None) and not np.array_equal(nucleus_label, cell_label) : Viewer.add_labels(cell_label, scale=scale, opacity= 0.2, blending= 'additive')
    if type(nucleus_label) != type(None) : Viewer.add_labels(nucleus_label, scale=scale, opacity= 0.2, blending= 'additive')
    
    #Adding widget
    if type(clusters) != type(None) :
        initialize_all_cluster_wizards(
            single_layer=single_layer,
            cluster_layer=cluster_layer
        )

        widget_clusterID = ClusterIDSetter(single_layer=single_layer, cluster_layer=cluster_layer)
        widget_cluster_merge =ClusterMerger(single_layer=single_layer, cluster_layer=cluster_layer)
        widget_cluster_updater = ClusterUpdater(
            single_layer=single_layer,
            cluster_layer=cluster_layer,
            default_cluster_radius= cluster_size,
            default_min_spot= min_spot_number,
            voxel_size=voxel_size
        )
        widget_cluster_creator = ClusterCreator(
            cluster_layer=cluster_layer,
            single_layer=single_layer
        )


        buttons_container = widgets.Container(widgets=[widget_clusterID.widget, widget_cluster_creator.widget], labels=False, layout='horizontal')
        updater_container = widgets.Container(widgets=[widget_cluster_updater.widget, widget_cluster_merge.widget], labels=False)
        tools_container = widgets.Container(
            widgets = [updater_container, buttons_container],
            labels=False,
        )
        Viewer.window.add_dock_widget(tools_container, name='SmallFish', area='left')

    Viewer.show(block=False)
    napari.run()

    new_spots = np.concatenate([
        single_layer.data,
        single_layer.features.loc[:,["cluster_id"]].to_numpy()
    ], axis=1).astype(int)

    if type(clusters) != type(None) :
        new_clusters = np.concatenate([
            cluster_layer.data,
            cluster_layer.features.loc[:,["spot_number","cluster_id"]].to_numpy()
        ],axis=1)

        new_cluster_radius = widget_cluster_updater.cluster_radius
        new_min_spot_number = widget_cluster_updater.min_spot

    else : 
        new_clusters = None
        new_cluster_radius = None
        new_min_spot_number = None

    return new_spots, new_clusters,  new_cluster_radius, new_min_spot_number


# Segmentation
def show_segmentation(
        nuc_image : np.ndarray,
        nuc_label : np.ndarray,
        cyto_image : np.ndarray = None,
        cyto_label : np.ndarray = None,
) :
    dim = nuc_image.ndim
    
    if type(cyto_image) != type(None) :
        if cyto_image.ndim != nuc_image.ndim : raise ValueError("Cyto and Nuc dimensions missmatch.")
        if type(cyto_label) == type(None) : raise ValueError("If cyto image is passed cyto label must be passed too.")

    if dim == 3 and nuc_label.ndim == 2 :
        nuc_label = np.repeat(
            nuc_label[np.newaxis],
            repeats= len(nuc_image),
            axis=0
        )
    if type(cyto_label) != type(None) :
        
        if type(cyto_image) == type(None) : raise ValueError("If cyto label is passed cyto image must be passed too.")
        
        if dim == 3 and cyto_label.ndim == 2 :
            cyto_label = np.repeat(
                cyto_label[np.newaxis],
                repeats= len(nuc_image),
                axis=0
            )

    #Init Napari viewer
    Viewer = napari.Viewer(ndisplay=2, title= 'Show segmentation', axis_labels=['z','y','x'] if dim == 3 else ['y', 'x'])
    
    # Adding nuclei
    nuc_signal_layer = Viewer.add_image(nuc_image, name= "nucleus signal", blending= 'additive', colormap='blue', contrast_limits=[nuc_image.min(), nuc_image.max()])
    nuc_label_layer = Viewer.add_labels(nuc_label, opacity= 0.6, name= 'nucleus_label',)
    nuc_label_layer.preserve_labels = True
    labels_layer_list = [nuc_label_layer]
    
    #Adding cytoplasm
    if (type(cyto_label) != type(None) and not np.array_equal(cyto_label, nuc_label) ) or (type(cyto_label) != type(None) and cyto_label.max() == 0): 
        Viewer.add_image(cyto_image, name= "cytoplasm signal", blending= 'additive', colormap='red', contrast_limits=[cyto_image.min(), cyto_image.max()])
        cyto_label_layer = Viewer.add_labels(cyto_label, opacity= 0.6, name= 'cytoplasm_label')
        cyto_label_layer.preserve_labels = True
        labels_layer_list += [cyto_label_layer]

    #Adding widget
    label_eraser = CellLabelEraser(labels_layer_list)
    label_picker = FreeLabelPicker(labels_layer_list)
    label_reseter = SegmentationReseter(labels_layer_list)
    changes_applier = ChangesPropagater(labels_layer_list)

    buttons_container = widgets.Container(widgets=[label_picker.widget, changes_applier.widget, label_reseter.widget], labels=False, layout='horizontal')
    tools_container = widgets.Container(
        widgets = [buttons_container, label_eraser.widget],
        labels=False,
    )
    Viewer.window.add_dock_widget(tools_container, name='SmallFish', area='left')

    #Launch Napari
    napari.run()

    new_nuc_label = Viewer.layers['nucleus_label'].data
    if 'cytoplasm_label' in Viewer.layers : 
        new_cyto_label = Viewer.layers['cytoplasm_label'].data
    else : new_cyto_label = new_nuc_label

    return new_nuc_label, new_cyto_label

def threshold_selection(
        image : np.ndarray,
        filtered_image : np.ndarray,
        threshold_slider,
        voxel_size : tuple,
        ) :
    
    """
    To view code for spot selection have a look at magicgui instance created with `detection._create_threshold_slider` which is then passed to this napari wrapper as 'threshold_slider' argument.
    """
    
    Viewer = napari.Viewer(title= "Small fish - Threshold selector", ndisplay=2, show=True)
    scale = compute_anisotropy_coef(voxel_size)
    Viewer.add_image(
        data= image,
        contrast_limits= [image.min(), image.max()],
        name= "raw signal",
        colormap= 'green',
        scale= scale,
        blending= 'additive'
    )
    Viewer.add_image(
        data= filtered_image,
        contrast_limits= [filtered_image.min(), filtered_image.max()],
        colormap= 'gray',
        scale=scale,
        blending='additive'
    )

    Viewer.window.add_dock_widget(threshold_slider, name='threshold_selector')
    threshold_slider() #First occurence with auto or entered threshold.
    
    napari.run()

    spots = Viewer.layers['single spots'].data.astype(int)
    if len(spots) == 0 :
        pass
    else :
        threshold = Viewer.layers['single spots'].properties.get('threshold')[0]

    return spots, threshold


"""
Submodule handling all parameters check, asserting functions and pipeline will be able to run.
"""

import os
import czifile as czi
import bigfish.stack as stack
import numpy as np
import FreeSimpleGUI as sg

from ..pipeline._preprocess import check_integrity, convert_parameters_types, ParameterInputError, _check_segmentation_parameters
from ..pipeline.segmentation import _cast_segmentation_parameters

def check_file(filename:str) :

    if filename.endswith('.czi') :
        image = czi.imread(filename)
    else :
        image = stack.read_image(filename)

    image = np.squeeze(image)

    return image.shape

def sanity_check(
        filename_list: list, 
        batch_folder : str,
        window : sg.Window, 
        progress_bar: sg.ProgressBar,
        ) :
    
    filenumber = len(filename_list)
    if filenumber == 0 :
        print("No file to check")
        progress_bar.update(current_count= 0, bar_color=('gray','gray'))
        return None
    else :
        print("{0} files to check".format(filenumber))
        progress_bar.update(current_count=0, max= filenumber)
        ref_shape = check_file(batch_folder + '/' + filename_list[0])

        print("Starting sanity check. This could take some time...")
        for i, file in enumerate(filename_list) :
            progress_bar.update(current_count= i+1, bar_color=('green','gray'))
            shape = check_file(batch_folder + '/' + file)

            if len(shape) != len(ref_shape) : #then dimension missmatch
                print("Different number of dimensions found : {0}, {1}".format(len(ref_shape), len(shape)))
                progress_bar.update(current_count=filenumber, bar_color=('red','black'))
                window= window.refresh()
                break

            window= window.refresh()

        print("Sanity check completed.")
        return None if len(shape) != len(ref_shape) else shape

def check_channel_map_integrity(
        maping:dict, 
        shape: tuple,
        expected_dim : int
        ) :
    
    #Check integrity
    channels_values = np.array(list(maping.values()), dtype= int)
    total_channels = len(maping)
    unique_channel = len(np.unique(channels_values))
    res= True

    if expected_dim != total_channels :
        sg.popup("Image has {0} dimensions but {1} were mapped.".format(expected_dim, total_channels))
        res = False
    if total_channels != unique_channel :
        sg.popup("{0} channel(s) are not uniquely mapped.".format(total_channels - unique_channel))
        res = False
    if not all(channels_values < len(shape)):
        sg.popup("Channels values out of range for image dimensions.\nPlease select dimensions from {0}".format(list(range(len(shape)))))
        res = False

    return res

def check_segmentation_parameters(
        values,
        shape,
        is_multichannel,
) :
    values = _cast_segmentation_parameters(values=values)
    try :
        _check_segmentation_parameters(
            user_parameters=values,
            shape=shape,
            is_multichannel=is_multichannel
        )
    except ParameterInputError as e: 
        segmentation_is_ok = False
        sg.popup_error(e)
    else :
        segmentation_is_ok = True
    
    return segmentation_is_ok, values

def check_detection_parameters(
        values,
        do_dense_region_deconvolution,
        do_clustering,
        is_multichannel,
        is_3D,
        map_,
        shape
) :
    
    values['dim'] = 3 if is_3D else 2
    values = convert_parameters_types(values)
    try :
        check_integrity(
            values=values,
            do_dense_region_deconvolution=do_dense_region_deconvolution,
            do_clustering=do_clustering,
            multichannel=is_multichannel,
            segmentation_done=None,
            map_=map_,
            shape=shape
        )
    except ParameterInputError as e: 
        detection_is_ok = False
        sg.popup_error(e)
    else :
        detection_is_ok = True
    
    return detection_is_ok, values

def check_output_parameters(values) :
    is_output_ok = True

    #Output folder
    output_folder = values.get('output_folder')
    if not os.path.isdir(output_folder) :
        sg.popup("Incorrect output folder selected")
        is_output_ok=False

    #Batch name
    original_name = values['batch_name']
    loop=1
    values['batch_name'] = values['batch_name'].replace(' ','_')
    while os.path.isdir(output_folder + '/' + values['batch_name']) :
        values['batch_name'] = original_name + '_{0}'.format(loop)
        loop+=1
    if len(values['batch_name']) == 0 : is_output_ok = False

    #extension
    if values['csv'] or values['xlsx'] or values['feather'] :
        pass
    else :
        sg.popup("Select at least one data format for output.")
        is_output_ok=False


    return is_output_ok, values


"""
Submodule containing custom class for napari widgets
"""
import numpy as np
import pandas as pd
import bigfish.detection as detection
from napari.layers import Labels, Points
from magicgui import magicgui

from abc import ABC, abstractmethod
from typing import Tuple

class NapariWidget(ABC) :
    """
    Common super class for custom widgets added to napari interface during run
    Each sub class as a specific function, but the widget can be acess with attribute .widget
    """
    def __init__(self):
        self.widget = self._create_widget()

    @abstractmethod
    def _create_widget(self) :
        """
        This should return a widget you can add to the napari (QWidget)
        """
        pass

class ClusterWidget(NapariWidget) :
    """
    Widget for clusters interaction are all init with cluster_layer and single_layer
    """
    def __init__(self, cluster_layer : Points, single_layer : Points):
        self.cluster_layer = cluster_layer
        self.single_layer = single_layer
        super().__init__()

class ClusterWizard(ABC) :
    """
    Common super class for all classes that will interact on single layer and cluster layer to synchronise them or modify their display.
    Their action is started through 'start_listening' method.
    To register them in CLUSTER_WIZARD they should only take single_layer and cluster_layer as arguments
    """

    def __init__(self, single_layer : Points, cluster_layer : Points):
        self.single_layer = single_layer
        self.cluster_layer = cluster_layer
        self.start_listening()
    
    def start_listening(self) :
        """
        This activate the class function. Returns None
        """
        pass


CLUSTER_WIZARDS = []
def register_cluster_wizard(cls):
    """
    Helper to register all clusters related class
    """
    CLUSTER_WIZARDS.append(cls)
    return cls

def initialize_all_cluster_wizards(single_layer: Points, cluster_layer: Points):
    """
    Initialize all wizards for cluster interaction in Napari
    """
    return [
        cls(single_layer, cluster_layer)
        for cls in CLUSTER_WIZARDS
    ]


class CellLabelEraser(NapariWidget) :
    """
    Widget for deleting cells from multiple label layers in a Napari viewer.
    """
    def __init__(self, label_list: 'list[Labels]'):
        self.label_list = label_list
        if len(self.label_list) == 0 : raise ValueError("Empty label list")
        for label_layer in self.label_list :
            label_layer.events.selected_label.connect((self, 'update'))
        super().__init__()

    def update(self, event) :
        layer : Labels = event.source
        new_label = layer.selected_label
        self.widget.label_number.value = new_label
        self.widget.update()
    
    def _create_widget(self) :
        @magicgui(
                call_button="Delete cell",
                auto_call=False
                )
        def label_eraser(label_number: int) -> None :

            for i, label in enumerate(self.label_list) :
                self.label_list[i].data[label.data == label_number] = 0
                label.refresh()

        return label_eraser


class FreeLabelPicker(NapariWidget) :
    """
    This widget gives the user a free label number
    """
    def __init__(self, label_list : 'list[Labels]'):
        self.label_list = label_list
        if len(self.label_list) == 0 : raise ValueError("Empty label list")
        super().__init__()
    
    def _create_widget(self) :
        @magicgui(
            call_button="Pick free label",
            auto_call=False
        )
        def label_pick()->None :
            max_list = [label_layer.data.max() for label_layer in self.label_list]
            new_label = max(max_list) + 1
            for label_layer in self.label_list :
                label_layer.selected_label = new_label
                label_layer.refresh()

        return label_pick


class SegmentationReseter(NapariWidget) :
    """
    This widget reset the segmentation mask as it used to be when iniating the instance
    """
    def __init__(self, label_list: 'list[Labels]'):
        self.label_list = label_list
        if len(self.label_list) == 0 : raise ValueError("Empty label list")
        self.save = self._get_save()
        super().__init__()
        
    
    def _get_save(self, label_list : 'list[Labels]') :
        return [label.data.copy() for label in label_list]

    def _create_widget(self) :
        @magicgui(
            call_button= 'Reset segmentation',
            auto_call=False,
        )
        def reset_segmentation() -> None:
            for save_data, layer in zip(self.save, self.label_list) :
                layer.data = save_data.copy()
                layer.refresh()

        return reset_segmentation

class ChangesPropagater(NapariWidget) :
    """
    Apply the changes across the vertical direction (Zstack) if confling values are found for a pixel, max label is kept.
    """
    def __init__(self, label_list):
        self.label_list = label_list
        super().__init__()

    def _create_widget(self) :
        @magicgui(
            call_button='Apply changes',
            auto_call=False,
        )
        def apply_changes() -> None:
            for layer in self.label_list :
                slices = layer.data.shape[0]
                layer_2D = np.max(layer.data, axis=0)
                layer.data = np.repeat(layer_2D[np.newaxis], slices, axis=0)
                layer.refresh()
        return apply_changes

class ClusterIDSetter(ClusterWidget) :
    """
    Allow user to set selected single spots to chosen cluster_id
    """
    def __init__(self, single_layer : Points, cluster_layer : Points):
        super().__init__(cluster_layer, single_layer)

    def _create_widget(self):

        @magicgui(
                call_button= "Set cluster ID",
                auto_call= False,
                cluster_id= {'min' : -1},
        )
        def set_cluster_id(cluster_id : int) :
            if cluster_id == -1 or cluster_id in self.cluster_layer.features['cluster_id'] :
                spots_selection = list(self.single_layer.selected_data)
                cluster_id_in_selection = list(self.single_layer.features.loc[spots_selection,["cluster_id"]].to_numpy().flatten()) + [cluster_id]
                self.single_layer.features.loc[spots_selection,["cluster_id"]] = cluster_id

                for cluster_id in np.unique(cluster_id_in_selection): # Then update number of spots in cluster
                    if cluster_id == -1 : continue
                    new_spot_number = len(self.single_layer.features.loc[self.single_layer.features['cluster_id'] == cluster_id])
                    self.cluster_layer.features.loc[self.cluster_layer.features['cluster_id'] == cluster_id, ["spot_number"]] = new_spot_number
                self.cluster_layer.events.features()

            self.cluster_layer.selected_data.clear()

        return set_cluster_id

class ClusterMerger(ClusterWidget) :
    """
    Merge all selected clusters by replacing cluster ids of all clusters and belonging points with min for cluster id.
    """
    def __init__(self, cluster_layer, single_layer):
        super().__init__(cluster_layer, single_layer)
    
    
    def _create_widget(self):

        @magicgui(
            call_button="Merge Clusters",
            auto_call=False
        )
        def merge_cluster()-> None :
            selected_clusters = list(self.cluster_layer.selected_data)
            if len(selected_clusters) == 0 : return None

            selected_cluster_ids = self.cluster_layer.features.loc[selected_clusters,['cluster_id']].to_numpy().flatten()
            new_cluster_id = selected_cluster_ids.min()

            #Dropping selected clusters
            self.cluster_layer.data = np.delete(self.cluster_layer.data, selected_clusters, axis=0)

            #Updating spots
            belonging_spots = self.single_layer.features.loc[self.single_layer.features['cluster_id'].isin(selected_cluster_ids)].index
            self.single_layer.features.loc[belonging_spots, ["cluster_id"]] = new_cluster_id

            #Creating new cluster
            centroid = list(self.single_layer.data[belonging_spots].mean(axis=0).round().astype(int))
            spot_number = len(belonging_spots)
            self.cluster_layer.data = np.append(
                self.cluster_layer.data,
                [centroid],
                axis=0
            )

            last_index = len(self.cluster_layer.data) - 1
            self.cluster_layer.features.loc[last_index, ['cluster_id']] = new_cluster_id
            self.cluster_layer.features.loc[last_index, ['spot_number']] = spot_number

            self.cluster_layer.selected_data.clear()
            self.cluster_layer.refresh()

        return merge_cluster




class ClusterUpdater(NapariWidget) :
    """
    Relaunch clustering algorithm taking into consideration new spots, new clusters and deleted clusters.
    """
    def __init__(
            self, 
            single_layer : Points, 
            cluster_layer : Points, 
            default_cluster_radius : int, 
            default_min_spot : int,
            voxel_size : 'tuple[int]'
            ):
        self.single_layer = single_layer
        self.cluster_layer = cluster_layer
        self.cluster_radius = default_cluster_radius
        self.min_spot = default_min_spot
        self.voxel_size = voxel_size
        super().__init__()

    def _compute_clusters(
            self, 
            cluster_radius : int, 
            min_spot : int
            ) -> Tuple[np.ndarray, np.ndarray, dict, dict] :
        """
        Compute clusters using bigfish detection.detect_clusters and seperate coordinates from features.
        """
        
        clustered_spots, clusters = detection.detect_clusters(
            voxel_size=self.voxel_size,
            spots= self.single_layer.data,
            radius=cluster_radius,
            nb_min_spots= min_spot
        )

        clusters_coordinates = clusters[:,:-2]
        clusters_features = {
            "spot_number" : clusters[:,-2],
            "cluster_id" : clusters[:,-1],
        }

        spots_coordinates = clustered_spots[:,:-1]
        spots_features = {
            "cluster_id" : clustered_spots[:,-1]
        }

        return clusters_coordinates, spots_coordinates, clusters_features, spots_features

    def _update_layers(
            self, 
            clusters_coordinates : np.ndarray, 
            spots_coordinates : np.ndarray, 
            clusters_features : dict, 
            spots_features : dict
            ) -> None  :
        """
        Update Points layers inside napari viewer.
        """
        
        #Modify layers
        self.single_layer.data = spots_coordinates
        self.cluster_layer.data = clusters_coordinates
        self.single_layer.features.loc[:,["cluster_id"]] = spots_features['cluster_id']
        self.cluster_layer.features.loc[:,["cluster_id"]] = clusters_features['cluster_id']
        self.cluster_layer.features.loc[:,["spot_number"]] = clusters_features['spot_number']

        self.cluster_layer.selected_data.clear()
        self.single_layer.refresh()
        self.cluster_layer.refresh()

        

    def _create_widget(self):

        @magicgui(
                call_button= "Relaunch Clustering",
                auto_call= False
        )
        def relaunch_clustering(
            cluster_radius : int = self.cluster_radius,
            min_spot : int = self.min_spot,
        ) :
            clusters_coordinates, spots_coordinates, clusters_features, spots_features = self._compute_clusters(cluster_radius=cluster_radius, min_spot=min_spot)
            self._update_layers(clusters_coordinates, spots_coordinates, clusters_features, spots_features )
            self.cluster_radius = cluster_radius
            self.min_spot = min_spot

        return relaunch_clustering

class ClusterCreator(ClusterWidget) :
    """
    Create a cluster containing all and only selected spots located at the centroid of selected points.
    """
    def __init__(self, cluster_layer, single_layer):
        super().__init__(cluster_layer, single_layer)

    def _create_widget(self):

        @magicgui(
                call_button= "Create Cluster",
                auto_call=False
        )
        def create_foci() -> None :
            selected_spots_idx = pd.Index(list(self.single_layer.selected_data))
            free_spots_idx : pd.Index = self.single_layer.features.loc[self.single_layer.features['cluster_id'] == -1].index
            selected_spots_idx = selected_spots_idx[selected_spots_idx.isin(free_spots_idx)]

            spot_number = len(selected_spots_idx)
            if spot_number == 0 :
                print("To create a cluster please select at least 1 spot")
            else :
                
                #Foci creation
                spots_coordinates = self.single_layer.data[selected_spots_idx]
                new_cluster_id = self.cluster_layer.features['cluster_id'].max() + 1
                centroid = list(spots_coordinates.mean(axis=0).round().astype(int))

                self.cluster_layer.data = np.concatenate([
                    self.cluster_layer.data,
                    [centroid]
                ], axis=0)
                
                last_index = len(self.cluster_layer.data) - 1
                self.cluster_layer.features.loc[last_index, ['cluster_id']] = new_cluster_id
                self.cluster_layer.features.loc[last_index, ['spot_number']] = spot_number

                #Update spots cluster_id
                self.single_layer.features.loc[selected_spots_idx,["cluster_id"]] = new_cluster_id
        
        return create_foci

@register_cluster_wizard
class ClusterInspector :
    """
    Listen to event on cluster layer to color spots belonging to clusters in green
    """
    def __init__(self, single_layer : Points, cluster_layer : Points):
        self.single_layer = single_layer
        self.cluster_layer = cluster_layer
        self.start_listening()

    def reset_single_colors(self) -> None:
        self.single_layer.face_color = [0,0,0,0] #transparent
        self.single_layer.refresh()

    def start_listening(self) :

        def color_single_molecule_in_foci() -> None:
            self.reset_single_colors()
            selected_cluster_indices = self.cluster_layer.selected_data
            for idx in selected_cluster_indices :
                selected_cluster = self.cluster_layer.features.at[idx,"cluster_id"]
                belonging_single_idex = self.single_layer.features.loc[self.single_layer.features['cluster_id'] == selected_cluster].index.to_numpy()
                self.single_layer.face_color[belonging_single_idex] = [0,1,0,1] #Green
                self.single_layer.refresh()

        self.cluster_layer.selected_data.events.items_changed.connect(color_single_molecule_in_foci)

@register_cluster_wizard
class ClusterEraser(ClusterWizard) :
    """
    When a foci is deleted, update spots feature table accordingly.
    """

    def __init__(self, single_layer, cluster_layer):
        super().__init__(single_layer, cluster_layer)

    def start_listening(self):
        self.original_remove_selected = self.cluster_layer.remove_selected
    
        def remove_selected_cluster() :
            selected_cluster = self.cluster_layer.selected_data
            for cluster_idx in selected_cluster : #First we update spots data
                cluster_id = self.cluster_layer.features.at[cluster_idx, "cluster_id"]
                self.single_layer.features.loc[self.single_layer.features['cluster_id'] == cluster_id, ['cluster_id']] = -1
            
            self.original_remove_selected() # Then we launch the usual napari method
        
        self.cluster_layer.remove_selected = remove_selected_cluster

@register_cluster_wizard
class ClusterAdditionDisabler(ClusterWizard) :
    """
    Remove the action when user uses points addition tool for Foci, forcing him to use the FociCreator tool to add new cluster.
    """

    def __init__(self, single_layer, cluster_layer):
        super().__init__(single_layer, cluster_layer)
    
    def start_listening(self):

        def print_excuse(*args, **kwargs):
            print("Spot addition is disabled for cluster layer. Use the foci creation tool below after selecting spots you want to cluster")

        self.cluster_layer.add = print_excuse

@register_cluster_wizard
class SingleEraser(ClusterWizard) :
    """
    When a single is deleted, update clusters feature table accordingly
    """

    def __init__(self, single_layer, cluster_layer):
        super().__init__(single_layer, cluster_layer)

    def start_listening(self):
        self._origin_remove_single = self.single_layer.remove_selected

        def delete_single(*args, **kwargs) :
            selected_single_idx = list(self.single_layer.selected_data)
            modified_cluster_ids = self.single_layer.features.loc[selected_single_idx, ["cluster_id"]].to_numpy().flatten()

            print(np.unique(modified_cluster_ids, return_counts=True))
            for cluster_id, count in zip(*np.unique(modified_cluster_ids, return_counts=True)): # Then update number of spots in cluster
                    if cluster_id == -1 : continue
                    new_spot_number = len(self.single_layer.features.loc[self.single_layer.features['cluster_id'] == cluster_id]) - count #minus number of spot with this cluster id we remove
                    print("new spot number : ", new_spot_number)
                    print('target cluster id : ', cluster_id)
                    self.cluster_layer.features.loc[self.cluster_layer.features['cluster_id'] == cluster_id, ["spot_number"]] = new_spot_number
            self._origin_remove_single()
            self.cluster_layer.events.features()
        
        self.single_layer.remove_selected = delete_single


@register_cluster_wizard
class ClusterCleaner(ClusterWizard) :
    """
    Deletes clusters if they drop to 0 single molecules.
    """

    def __init__(self, single_layer, cluster_layer):
        super().__init__(single_layer, cluster_layer)

    def start_listening(self):

        def delete_empty_cluster() :
            drop_idx = self.cluster_layer.features[self.cluster_layer.features['spot_number'] == 0].index
            print("drop_idx : ",drop_idx)
            
            if len(drop_idx) > 0 :
                print("Removing {} empty cluster(s)".format(len(drop_idx)))
                self.cluster_layer.data = np.delete(self.cluster_layer.data, drop_idx, axis=0)
                self.cluster_layer.refresh()

        self.cluster_layer.events.features.connect(delete_empty_cluster)
        


import sys, subprocess, re
from small_fish_gui import __version__

AVAILABLE_ARGUMENTS = {
    ('-v','--v','--version') : "Prompt the software version.",
    ('--launch', '-l') : "Launch small fish gui, equivalent to no arguments.",
    ('-h', '--help', '--h') : "Prompt this help menu."
}


def main():
    import small_fish_gui.pipeline.main

def _get_version() :
    return __version__

def is_last_version() :
    
    query = subprocess.run([sys.executable, '-m', 'pip', 'index', 'versions', 'small_fish_gui'], capture_output=True, text=True)
    all_version = query.stdout.split(',')
    latest_version = all_version[0]
    regex = r"\d+\.\d+\.\d+"
    latest = re.findall(regex, latest_version)
    
    current_version = _get_version()
    
    if len(latest) == 0 : 
        return current_version
    else :
        return current_version == latest[-1]

if __name__ == "__main__":

    if not is_last_version() :
        print("A new version of Small Fish is available. To update close small fish and type :\npip install --upgrade small_fish_gui")

    arguments = sys.argv

    if len(arguments) > 1 :
        if arguments[1] in ['-v','--v','--version'] :
            print(_get_version())
            quit()
        elif arguments[1] in ['--launch', '-l'] :
            pass
        elif arguments[1] in ['-h', '--help', '--h'] :
            for key, help in AVAILABLE_ARGUMENTS.items() :
                print(f"{key} : {help}")
            quit()
        else :
            print(f"Incorrect argument : {arguments}, to launch small fish don't pass any argument or pick amongst {AVAILABLE_ARGUMENTS.keys()}")

    sys.exit(main())


"""
Module containing main pipeline for user mode as well as calls for pipeline functions.
"""

from ._preprocess import reorder_shape
from ._preprocess import reorder_image_stack
from ._preprocess import prepare_image_detection
from ._preprocess import convert_parameters_types

from .segmentation import launch_segmentation
from .segmentation import _cast_segmentation_parameters
from .segmentation import cell_segmentation
from .segmentation import plot_segmentation

from .detection import launch_detection
from .detection import launch_features_computation
from .detection import launch_cell_extraction
from .detection import get_nucleus_signal
from .detection import output_spot_tiffvisual

from .spots import launch_spots_extraction



#Add keys hinting to user_parameters instance keys.

from typing import TypedDict, Tuple
from numpy import ndarray
    
class pipeline_parameters(TypedDict) :
            """
            At run time is a regular dict instance, this class is used for keys hinting
            """
            alpha : float
            beta : float
            channel_to_compute : int
            cluster_size : int
            do_cluster_computation : bool
            do_dense_regions_deconvolution : bool
            do_spots_excel : bool
            do_spots_feather : bool
            do_spots_csv : bool
            dim : int
            filename : str
            gamma : float
            image_path : str
            image : ndarray
            show_interactive_threshold_selector : bool
            log_kernel_size : Tuple[float,float,float]
            log_kernel_size_x : float
            log_kernel_size_y : float
            log_kernel_size_z : float
            min_number_of_spots : int
            minimum_distance : Tuple[float,float,float]
            minimum_distance_x : float
            minimum_distance_y : float
            minimum_distance_z : float
            is_3D_stack : bool
            is_multichannel : bool
            show_napari_corrector : bool
            nucleus_channel_signal : int
            reodered_shape : Tuple[int,int,int,int,int]
            do_segmentation : bool
            segmentation_done : bool
            shape : Tuple[int,int,int,int,int]
            spots_extraction_folder : str
            spots_filename : str
            spot_size : Tuple[int,int,int]
            spot_size_x : int
            spot_size_y : int
            spot_size_z : int
            threshold : int
            threshold_penalty : int
            time_stack : None
            time_step : None
            voxel_size : Tuple[float,float,float]
            voxel_size_x : float
            voxel_size_y : float
            voxel_size_z : float


import FreeSimpleGUI as sg
import numpy as np

WAITING_TEXT = [
    'Running...',
    'Computing science...',
    'Preparing some good results...',
    'A good day for science...'
]

def add_default_loading(funct) :
    def inner(*args,**kwargs) :

        hide_loading = kwargs.get("hide_loading")
        if 'hide_loading' in kwargs : del kwargs['hide_loading']

        if not hide_loading :
            random_text = np.random.randint(0,len(WAITING_TEXT))
            waiting_layout = [
            [sg.Text(WAITING_TEXT[random_text], font= '10')]
            ]
            window = sg.Window(
            title= 'small_fish',
            layout= waiting_layout,
            grab_anywhere= True,
            finalize=True
        )

            window.read(timeout= 30, close= False)
            try :
                return funct(*args, **kwargs)
            finally :
                window.close()

        else : 
            return funct(*args, **kwargs)

    return inner

